/* ########################################################################

				While.c

   File: While.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/While.c
   Description: 
   Created: Tue Feb 21 12:47:04 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 12:47:05 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



#include "While_.h"
#include "stack.h"
#include "control.h"

/* WHILE */

static inherit_instruction_function_vector
  while_inherit_instruction_function_vector;

While * While__While(test, body, hasbreakcont)
     Instruction * test;
     Instruction * body;
     int hasbreakcont;
{
  if (GetExprType(test) == Type_Void) {
    Error("while (void) ..\n", 1);
    return (While *) ERROR;
  }
  else {
    While * this = (While *) Malloc(sizeof(While));
    
    Instruction_Constructor
      (while_inherit_instruction_function_vector, this, Type_Void);
    
    this->_test = test;
    
    this->_body = body;
    this->_hasbreakcont = hasbreakcont;
    
    return this;
  }
}


static void While__Delete(this)
     While * this;
{
  Delete(this->_test);
  if (this->_body)
    Delete(this->_body);

  free(this);
}

static Object While__Eval(this)
     While * this;
{
  Instruction * test = this->_test;
  Instruction * body = this->_body;
  
  if (body)
    if (this->_hasbreakcont) {
      Object * stackpointer = Stack_Pointer;
      JMP_TYPE jmptype;
      jmp_buf * spcurrent_env = pCurrent_Env;
      jmp_buf current_env;
      
      while(Eval(test)) {
	switch (jmptype = (JMP_TYPE) setjmp(current_env)) {
	case NO_JMP :
	  pCurrent_Env = &current_env;
	  Eval(body);
	  break;
	case JMP_CONTINUE:
	  Stack_Pointer = stackpointer;
	  goto etiq_cont;
	case JMP_BREAK:
	  Stack_Pointer = stackpointer;
	  goto etiq_break;
	default:
	  Stack_Pointer = stackpointer;
	  pCurrent_Env = spcurrent_env;
	  longjmp(*spcurrent_env, jmptype);
	}
      etiq_cont: ;
	pCurrent_Env = spcurrent_env;	/* pour le test */
      }
    etiq_break:
      pCurrent_Env = spcurrent_env;
    }
    else
      while(Eval(test)) Eval(body);
  else
    while(Eval(test)) ;

  return 0;
}


/* DO WHILE */

static inherit_instruction_function_vector
  dowhile_inherit_instruction_function_vector;

While * DoWhile__DoWhile(test, body, hasbreakcont)
     Instruction * test;
     Instruction * body;
     int hasbreakcont;
{
  if (GetExprType(test) == Type_Void) {
    Error("do .. while(void)\n", 1);
    return (While *) ERROR;
  }
  else {
    While * this = (While *) Malloc(sizeof(While));
    
    Instruction_Constructor
      (dowhile_inherit_instruction_function_vector, this, Type_Void);
    
    this->_test = test;
    
    this->_body = body;
    this->_hasbreakcont = hasbreakcont;
    
    return this;
  }
}


static void DoWhile__Delete(this)
     While * this;
{
  if (this->_body)
    Delete(this->_body);
  Delete(this->_test);

  free(this);
}

static Object DoWhile__Eval(this)
     While * this;
{
  Instruction * test = this->_test;
  Instruction * body = this->_body;
  
  if (body)
    if (this->_hasbreakcont) {
      Object * stackpointer = Stack_Pointer;
      JMP_TYPE jmptype;
      jmp_buf * spcurrent_env = pCurrent_Env;
      jmp_buf current_env;
      
      do {
	switch (jmptype = (JMP_TYPE) setjmp(current_env)) {
	case NO_JMP :
	  pCurrent_Env = &current_env;
	  Eval(body);
	  break;
	case JMP_CONTINUE:
	  Stack_Pointer = stackpointer;
	  goto etiq_cont;
	case JMP_BREAK:
	  Stack_Pointer = stackpointer;
	  goto etiq_break;
	default:
	  Stack_Pointer = stackpointer;
	  pCurrent_Env = spcurrent_env;
	  longjmp(*spcurrent_env, jmptype);
	}
      etiq_cont: 
	pCurrent_Env = spcurrent_env;	/* pour le test */
      } while (Eval(test));
    etiq_break:
      pCurrent_Env = spcurrent_env;
    }
    else
      do Eval(body); while(Eval(test));
  else
    while(Eval(test)) ;

  return 0;
}


void Init_While()
{
  while_inherit_instruction_function_vector._Eval = (EvalFct) While__Eval;
  while_inherit_instruction_function_vector._Delete = (DeleteFct) While__Delete;
  
  dowhile_inherit_instruction_function_vector._Eval = (EvalFct) DoWhile__Eval;
  dowhile_inherit_instruction_function_vector._Delete = (DeleteFct) DoWhile__Delete;
}
