/* ########################################################################

				 type.c

   File: type.c
   Path: /home/fournigault/c/X11/xcoral-2.33/Smac/type.c
   Description: 
   Created: Tue Feb 21 13:02:16 MET 1995
   Author: Bruno Pages
   Modified: Tue Feb 21 13:02:17 MET 1995
   Last maintained by: Bruno Pages

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Bruno Pages

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */



#include <stdio.h>

#include "type.h"
#include "error.h"
#include "list.h"
#include "memo.h"
#include "decl.h"
#include "Const_.h"

#include "decl.h" /**/

Instruction * make_typedef(l, del)
    List * l;
    int del;
{
  free(l);

  if (!del)
    Internal_Error("make_typedef called out of del", 0);
  
  return 0;
}


/* Retourne le type et detruit l */

char * asterisk = (char *) "asterisk";

static FCT(Type *, pointeur_ou_tableau_de_fonction,(List *, List **));
static FCT(Type *, pointeur_ou_tableau_de_tableau,(List *, List **));

Type * abstract_type(l)
    List * l;
{
  Type * result = 0;	/* initialize to not have warning */
  List * p;

  for (p = l; p && (p->fct != preserve_list); p = p->next);

  if (p) {
    /* Description de type complexe :

       int (*)()     : & ( int .() )      : pt sur fonc
       int (*[1])()  : .[1] = &(int . ()) : vect de pt de fonc
       
       int (*)[1]    : & ( int . [1] )    : pt sur vect d'int
       int (*[1])[2] : .[1] = &(int .[2]) : vect de pt de vect d'int

       Note : les fonctions retournant un type complexe ont ete traitees par
              l'appelant, il n'y a donc pas ici de cas tel que  int (*f())()
    */

    if ((((List *) p->info)->fct != just_a_mark) ||
	(((List *) p->info)->info != (Object) asterisk)){
      Internal_Error("I cannot understand the type description", 1);
      return 0;
    }
    
    if (p->next->fct == function_parameter_type_list) {
      result = pointeur_ou_tableau_de_fonction(p, &l);
    }

    else if (p->next->fct == array_dims) {
      result = pointeur_ou_tableau_de_tableau(p, &l);
      
      if (result == 0)
	return 0;
    }
    
    else {
      Internal_Error("I cannot understand the type description", 1);
      return 0;
    }
  }

  while (l) {
    List * suiv = l->next;
    if (l->fct == make_typedef) {
      /* un type builtin */
      result = (Type *) l->info;
      free(l);
      l = suiv;
    }
    else if ((l->fct == just_a_mark) && (l->info == (Object) asterisk)) {
      /* * */
      result = Type__Type(result, T_Pointer, sizeof(void *));
      free(l);
      l = suiv;
    }
    else if (l->fct == array_dims) {
      List * d = (List *) l->info;

      while (d) {
	List * prev = d->previous;

	if (! d->fct) {
	  Internal_Error("[] not yet implemented", 1);
	  free(d);
	  do {
	    suiv = l->next;
	    free(l);
	  } while ((l = suiv) != 0);
	  return 0;
	}
	else {
	  Instruction * i = MAKE(d);
	  
	  if (i == ERROR) {
	    do {
	      suiv = l->next;
	      free(l);
	    } while ((l = suiv) != 0);
	    return 0;
	  }
	  else {
	    int dim = Eval(i);
	    
	    Delete(i);
	    if (dim <= 0) {
	      sprintf(err_msg, "%d illegal array dimension", dim);
	      Error(err_msg, 1);
	      do {
		suiv = l->next;
		free(l);
	      } while ((l = suiv) != 0);
	      return 0;
	    }
	    result = Type__Type(result, T_Array, dim * Type__Sizeof(result));
	  }
	}

	d = prev;
      }
      free(l);
      l = suiv;
    }
    else {
      Error("bad type declaration", 1);
      do {
	suiv = l->next;
	free(l);
      } while ((l = suiv) != 0);
      return 0;
    }
  }

  return result;
}


static Type * pointeur_ou_tableau_de_fonction(p, l)
     List * p;
     List ** l;
{
  /*
       int (*)()     : & ( int .() )      : pt sur fonc
       int (*[1])()  : .[1] = &(int . ()) : vect de pt de fonc

       p pointe sur (*..), *l sur tout

       retourne le type fonction pointe, et change l en les []
  */
  
  TypeFunction * typefunction =
    (TypeFunction *) Malloc(sizeof(TypeFunction));
  List * params = (List *) p->next->info;
  Type ** argtypes;
  Type ** pargtypes;
  Type * result;

  typefunction->_nbre_arg = list_length(params);
  free(p->next);
  argtypes = pargtypes = (Type **)
    Malloc((typefunction->_nbre_arg) ? typefunction->_nbre_arg * sizeof(Type *) : 1);
  typefunction->_argtypes = argtypes;
  while (params) {
    /* ellipsis (...) actuellement ecarte dans lex */
    List * suite = params->next;
	
    *pargtypes++ = (Type *) ((Object) params->fct);
    if (params->info)				/* le nom de la var */
      free((void *) params->info);
    free(params);
    params = suite;
  }
  p->previous->next = 0;
  typefunction->_valtype = abstract_type(*l);
  if (typefunction->_valtype == 0) {
    free(argtypes);
    free(typefunction);
    free((void *) p->info);
    free(p);
    return 0;
  }
  result = Type__Type(typefunction, T_Function, sizeof(void **));
  {
    Type ** at = &TypeFunction__ArgType(Type__Function(result), 0);
    
    if (at != argtypes)
      /* function profile already exist */
      free(argtypes);
  }
  
  *l = ((List *) p->info)->next;		/* les [] s'ils existent */
  free((void *) p->info);
  free(p);

  return result;
}

static Type * pointeur_ou_tableau_de_tableau(p, l)
     List * p;
     List ** l;
{
  /*
       int (*)[1]    : & ( int . [1] )    : pt sur vect d'int
       int (*[1])[2] : .[1] = &(int .[2]) : vect de pt de vect d'int

       p pointe sur (*..), *l sur tout

       retourne le type tableau pointe, et change l en les []
  */

  Type * result;
  
  p->previous->next = p->next;			/* retire (*..) */
  result = abstract_type(*l);

  *l = (List *) p->info;

  return result;
}


/* Retourne une chaine de caract correspondant a un
   type pour s'en servir de clef d'une hash table */

char * encode_type(t)
     Type * t;
{
  if (Type__IsBuiltin(t))
    return Type__Name(t);

  if (Type__IsaFunction(t))
    return "<function>";
  
  if (Type__Pointed_Type(t) != Type_Char)
    return "<void *>";

  return "<char *>";
}


/* Sizeof 

   fct = make_sizeof_type
   info = Type * type
   */

Instruction * make_sizeof_type(l, del)
    List * l;
    int del;
{
  Instruction * result =
    (del) ? (Instruction *) 0
	  : (Instruction *) Int__Int(Type__Sizeof((Type *) l->info));

  free(l);

  return result;
}
