/* ########################################################################

			      class_dict.c

   File: class_dict.c
   Path: /home/fournigault/c/X11/xcoral-2.31/class_dict.c
   Description: 
   Created: Fri Jan 27 10:53:29 MET 1995
   Author: Dominique Leveque
   Modified: Fri Jan 27 10:53:30 MET 1995
   Last maintained by: Dominique Leveque

   RCS $Revision$ $State$
   

   ########################################################################

   Note: 

   ########################################################################

   Copyright (c) : Dominique Leveque

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */


#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdlib.h>

#include "result_types.h"
#include "file_dict.h"
#include "class_dict.h"
#include "browser_util.h"


/*------------------------------------------------------------------------------
//                         Le dictionnaire des classes
//------------------------------------------------------------------------------
*/

ClassRec*  class_dict[CLASS_DICT_SIZE];

int        class_count = 0;


/*------------------------------------------------------------------------------
   create the class
   warning : its _name if part of the object, never free it separately !
*/
ClassRec* create_class (class_name, decl)
  char* class_name;
  unsigned decl;
{
  ClassRec**  head;
  ClassRec*   current_class;
  int         x_size;
  
  get_head_Rec(class_name, class_dict, CLASS_DICT_SIZE, head);
  search_Rec(class_name, ClassRec, head, current_class);
  if (current_class == Null) {
    x_size = sizeof(ClassRec) + CLASS_PLENGTH + strlen(class_name) + 1;
    current_class = (ClassRec*) xmalloc(x_size);
    if (current_class != Null) {
      create_Rec(class_name, ClassRec, head, current_class, CLASS_PREFIX, CLASS_PLENGTH);
      current_class->_decl_file     = Null;
      current_class->_decl_line     = 0;
      current_class->_parents_list  = Null;
      current_class->_parents_count = 0;
      current_class->_methods_list  = Null;
      current_class->_attributes_list  = Null;
      current_class->_decl = decl;
      current_class->_hide = 0;
      current_class->_next_marked   = Null;
      class_count++;
    }
  }
  return(current_class);
}


/*------------------------------------------------------------------------------
*/
ClassRec* find_class(class_name)
  char* class_name;
{
	ClassRec**  head;
  ClassRec*   current_class;

	get_head_Rec(class_name, class_dict, CLASS_DICT_SIZE, head);
  search_Rec(class_name, ClassRec, head, current_class);
  return(current_class);
}


/*------------------------------------------------------------------------------
*/
void class_eraze_file(file_name)
    char* file_name;
{
  FileRec*   current_file;
  ClassRec ** pdict;
  
  if (file_name == 0)
    current_file = 0;
  else if ((current_file = find_file(file_name)) == 0)
    return;

  for (pdict = class_dict;
       pdict != &class_dict[CLASS_DICT_SIZE];
       pdict += 1) {
    ClassRec ** p_class = pdict;
    ClassRec * current_class;
    
    while ((current_class = *p_class) != Null) {
      /* manage class decl */
      if ((current_file == 0) || (current_class->_decl_file == current_file)) {
	ParentRec * parent;
	
	current_class->_decl_file = Null;
	current_class->_decl_line = 0;
	current_class->_parents_count = 0;
	
	/* remove parents */
	parent = current_class->_parents_list;
	current_class->_parents_list = 0;
	while (parent != Null) {
	  ParentRec * removed = parent;
	  
	  parent = parent->_next;
	  free(removed);
	}
      }
      
      /* manage methods */
      {
	MethodRec ** p_method = &current_class->_methods_list;
	MethodRec * current_method;
	
	while ((current_method = *p_method) != Null) {
	  if (current_method->_decl_file == current_file) {
	    current_method->_decl_file = Null;
	    current_method->_decl_line = 0;
	  }
	  if (current_method->_impl_file == current_file) {
	    current_method->_impl_file = Null;
	    current_method->_impl_line = 0;
	  }
	  if ((current_file == 0) || 
	      ((current_method->_impl_file == Null) &&
	       (current_method->_decl_file == Null))) {
	    /* remove it */
	    *p_method = current_method->_next;
	    free(current_method);
	  }
	  else
	    /* keep it */
	    p_method = &current_method->_next;
	}
      }
      /* manage attributs */
      {
	AttributeRec ** p_attr = &current_class->_attributes_list;
	AttributeRec * current_attr;

	while ((current_attr = *p_attr) != Null) {
	  if ((current_file == 0) || (current_attr->_decl_file == current_file)) {
	    /* remove it */
	    *p_attr = current_attr->_next;
	    free(current_attr);
	  }
	  else
	    /* keep it */
	    p_attr = &current_attr->_next;
	}
      }
      
      /* the class is removed only if its definition is empty */
      if ((current_class->_decl_file == Null) &&
	  (current_class->_methods_list == Null) &&
	  (current_class->_attributes_list == Null)) {
	/* remove it */
	*p_class = current_class->_next;
	free(current_class);
	class_count -= 1;
      }
      else
	/* keep it */
	p_class = &current_class->_next;
    }
  }
}


/*------------------------------------------------------------------------------
*/

void class_eraze_marked_files()
{
  //FileRec*   current_file;
  ClassRec ** pdict;
  
  for (pdict = class_dict;
       pdict != &class_dict[CLASS_DICT_SIZE];
       pdict += 1) {
    ClassRec ** p_class = pdict;
    ClassRec * current_class;
    
    while ((current_class = *p_class) != Null) {
      /* manage class decl */
      if ((current_class->_decl_file != Null) &&
	  (current_class->_decl_file->_name == &Mark)) {
	ParentRec * parent;
	
	current_class->_decl_file = Null;
	current_class->_decl_line = 0;
	current_class->_parents_count = 0;
	
	/* remove parents */
	parent = current_class->_parents_list;
	current_class->_parents_list = 0;
	while (parent != Null) {
	  ParentRec * removed = parent;
	  
	  parent = parent->_next;
	  free(removed);
	}
      }
      
      /* manage methods */
      {
	MethodRec ** p_method = &current_class->_methods_list;
	MethodRec * current_method;
	
	while ((current_method = *p_method) != Null) {
	  if ((current_method->_decl_file != Null) &&
	      (current_method->_decl_file->_name == &Mark)) {
	    current_method->_decl_file = Null;
	    current_method->_decl_line = 0;
	  }
	  if ((current_method->_impl_file != Null) &&
	      (current_method->_impl_file->_name == &Mark)) {
	    current_method->_impl_file = Null;
	    current_method->_impl_line = 0;
	  }
	  if ((current_method->_impl_file == Null) &&
	      (current_method->_decl_file == Null)) {
	    /* remove it */
	    *p_method = current_method->_next;
	    free(current_method);
	  }
	  else
	    /* keep it */
	    p_method = &current_method->_next;
	}
      }
      /* manage attributs */
      {
	AttributeRec ** p_attr = &current_class->_attributes_list;
	AttributeRec * current_attr;

	while ((current_attr = *p_attr) != Null) {
	  if ((current_attr->_decl_file != Null) &&
	      (current_attr->_decl_file->_name == &Mark)) {
	    /* remove it */
	    *p_attr = current_attr->_next;
	    free(current_attr);
	  }
	  else
	    /* keep it */
	    p_attr = &current_attr->_next;
	}
      }
      
      /* the class is removed only if its definition is empty */
      if ((current_class->_decl_file == Null) &&
	  (current_class->_methods_list == Null) &&
	  (current_class->_attributes_list == Null)) {
	/* remove it */
	*p_class = current_class->_next;
	free(current_class);
	class_count -= 1;
      }
      else
	/* keep it */
	p_class = &current_class->_next;
    }
  }
}
